import sys
import typing
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, fields, is_dataclass
from datetime import datetime
from decimal import Decimal
from typing import (
    Any,
    cast,
    ClassVar,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
)


from .annotations_tricks import (
    get_ClassVar_arg,
    get_Type_arg,
    is_ClassVar,
    is_NewType,
    is_Type,
    is_TypeLike,
    name_for_type_like,
)
from .constants import (
    BINDINGS_ATT,
    cache_enabled,
    DEPENDS_ATT,
    enable_type_checking,
    GENERIC_ATT2,
    GENERIC_BOUND_ATT,
    MakeTypeCache,
    PYTHON_36,
    ZuperTypingGlobals,
)
from .exceptions import ZTypeError, ZValueError
from .logging import logger
from .recursive_tricks import (
    find_typevars_inside,
    get_name_without_brackets,
    NoConstructorImplemented,
    replace_typevars,
    TypeLike,
)
from .subcheck import can_be_used_as2

X = TypeVar("X")


def as_tuple(x) -> Tuple:
    return x if isinstance(x, tuple) else (x,)


if PYTHON_36:  # pragma: no cover
    from typing import GenericMeta

    # noinspection PyUnresolvedReferences
    old_one = GenericMeta.__getitem__
else:
    old_one = None

if PYTHON_36:  # pragma: no cover
    # logger.info('In Python 3.6')
    class ZMeta(type):
        def __getitem__(self, *params):
            return ZenericFix.__class_getitem__(*params)


else:
    ZMeta = type


class ZenericFix(metaclass=ZMeta):
    if PYTHON_36:  # pragma: no cover

        def __getitem__(self, *params):
            # logger.info(f'P36 {params} {self}')
            if self is typing.Generic:
                return ZenericFix.__class_getitem__(*params)

            if self is Dict:
                K, V = params
                if K is not str:
                    from .my_dict import make_dict

                    return make_dict(K, V)

            # noinspection PyArgumentList
            return old_one(self, params)

    # noinspection PyMethodParameters
    @classmethod
    def __class_getitem__(cls0, params):
        # logger.info(f"ZenericFix.__class_getitem__ params = {params}")
        types0 = as_tuple(params)
        assert isinstance(types0, tuple)
        for t in types0:
            assert is_TypeLike(t), (t, types0)

        bound_att = types0
        name = "Generic[%s]" % ",".join(name_for_type_like(_) for _ in bound_att)
        gp = type(name, (GenericProxy,), {GENERIC_ATT2: types0})
        setattr(gp, GENERIC_ATT2, types0)
        setattr(gp, BINDINGS_ATT, {})
        setattr(gp, GENERIC_BOUND_ATT, bound_att)
        return gp


class StructuralTyping(type):
    def __subclasscheck__(self, subclass) -> bool:
        can = can_be_used_as2(subclass, self)
        return can.result

    def __instancecheck__(self, instance) -> bool:
        T = type(instance)
        if T is self:
            return True
        if not is_dataclass(T):
            return False
        i = super().__instancecheck__(instance)
        if i:
            return True
        res = can_be_used_as2(T, self)

        return res.result


class MyABC(StructuralTyping, ABCMeta):
    def __new__(mcs, name_orig: str, bases, namespace, **kwargs):
        try:
            from .debug_print_ import debug_print

            d = debug_print(
                dict(
                    mcs=mcs,
                    name_orig=name_orig,
                    bases=bases,
                    namespace=namespace,
                    kwargs=kwargs,
                )
            )
        except:
            d = "-"
        # logger.info(f"----\nCreating name: {name_orig}\n{d}")
        # logger.info('namespace: %s' % debug_print(namespace))
        # logger.info('bases: %s' % debug_print(bases))
        # if bases:
        #     logger.info('bases[0]: %s' % str(bases[0].__dict__))
        if GENERIC_ATT2 in namespace:
            spec = namespace[GENERIC_ATT2]
        elif bases and GENERIC_ATT2 in bases[0].__dict__:
            spec = bases[0].__dict__[GENERIC_ATT2]
        else:
            spec = {}
        if spec:
            name0 = get_name_without_brackets(name_orig)
            name = f"{name0}[%s]" % (",".join(name_for_type_like(_) for _ in spec))
        else:
            name = name_orig
        #
        # if GENERIC_BOUND_ATT in namespace:
        #     bound_att = namespace[GENERIC_BOUND_ATT]
        # elif bases and GENERIC_BOUND_ATT in bases[0].__dict__:
        #     bound_att = bases[0].__dict__[GENERIC_BOUND_ATT]
        # else:
        #     bound_att = ()

        # noinspection PyArgumentList
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)
        qn = cls.__qualname__.replace("." + name_orig, "." + name)

        setattr(cls, "__qualname__", qn)
        setattr(cls, "__module__", mcs.__module__)
        setattr(cls, GENERIC_ATT2, spec)
        setattr(cls, GENERIC_BOUND_ATT, spec)
        # setattr(cls, GENERIC_BOUND_ATT, getattr(mcs, GENERIC_BOUND_ATT, ()))
        return cls


if PYTHON_36:  # pragma: no cover

    class FakeGenericMeta(MyABC):
        def __getitem__(self, params2):
            types = getattr(self, GENERIC_ATT2, ())

            # logger.info(f'FakeGenericMeta {params2!r} {self}')
            # pprint('FakeGenericMeta.__getitem__', cls=cls, self=self,
            # params2=params2)
            types2 = as_tuple(params2)

            assert isinstance(types2, tuple), types2
            for t in types2:
                assert is_TypeLike(t), (t, types2)

            if types == types2:
                return self

            bindings: Dict[TypeVar, TypeVar] = {}
            for T, U in zip(types, types2):
                bindings[T] = U
                if T.__bound__ is not None and isinstance(T.__bound__, type):
                    if not issubclass(U, T.__bound__):
                        msg = (
                            f'For type parameter "{T.__name__}", expected a'
                            f'subclass of "{T.__bound__.__name__}", found {U}.'
                        )
                        raise ZTypeError(msg)

            return make_type(self, bindings)


else:
    FakeGenericMeta = MyABC


class GenericProxy(metaclass=FakeGenericMeta):
    @abstractmethod
    def need(self) -> None:
        """"""

    @classmethod
    def __class_getitem__(cls, params2) -> type:
        # logger.info(f"GenericProxy.__class_getitem__ params = {params2}")

        types = getattr(cls, GENERIC_ATT2, ())
        types2 = as_tuple(params2)

        if len(types) != len(types2):
            msg = "Cannot match type length"
            raise ZValueError(msg, types=types, types2=types2)

        bindings = {}

        for T, U in zip(types, types2):
            bindings[T] = U
            if T.__bound__ is not None and isinstance(T.__bound__, type):
                # logger.info(f"{U} should be usable as {T.__bound__}")
                # logger.info(
                #     f" issubclass({U}, {T.__bound__}) ="
                #     f" {issubclass(U, T.__bound__)}"
                # )
                if not issubclass(U, T.__bound__):
                    msg = (
                        f'For type parameter "{T.__name__}", expected a'
                        f'subclass of "{T.__bound__.__name__}", found @U.'
                    )
                    raise ZTypeError(msg, T=T, T_bound=T.__bound__, U=U)

        res = make_type(cls, bindings)
        from .monkey_patching_typing import remember_created_class

        remember_created_class(res, "__class_getitem__")
        return res


class Fake:
    symbols: dict
    myt: type

    def __init__(self, myt, symbols: dict):
        self.myt = myt
        n = name_for_type_like(myt)
        self.name_without = get_name_without_brackets(n)
        self.symbols = symbols

    def __getitem__(self, item: type) -> type:
        n = name_for_type_like(item)
        complete = f"{self.name_without}[{n}]"
        if complete in self.symbols:
            return self.symbols[complete]
        # noinspection PyUnresolvedReferences
        return self.myt[item]


def resolve_types(
    T, locals_=None, refs: Tuple = (), nrefs: Optional[Dict[str, Any]] = None
):
    if nrefs is None:
        nrefs = {}

    assert is_dataclass(T)
    # rl = RecLogger()

    if locals_ is None:
        locals_ = {}
    symbols = dict(locals_)

    for k, v in nrefs.items():
        symbols[k] = v

    others = getattr(T, DEPENDS_ATT, ())

    for t in (T,) + refs + others:
        n = name_for_type_like(t)
        symbols[n] = t
        # logger.info(f't = {t} n {n}')
        name_without = get_name_without_brackets(n)

        # if name_without in ['Union', 'Dict', ]:
        #     # FIXME please add more here
        #     continue
        if name_without not in symbols:
            symbols[name_without] = Fake(t, symbols)
        # else:
        #     pass

    for x in getattr(T, GENERIC_ATT2, ()):
        if hasattr(x, "__name__"):
            symbols[x.__name__] = x

    # logger.debug(f'symbols: {symbols}')
    annotations: Dict[str, TypeLike] = getattr(T, "__annotations__", {})

    for k, v in annotations.items():
        if not isinstance(v, str) and is_ClassVar(v):
            continue  # XXX
        v = cast(TypeLike, v)
        try:
            r = replace_typevars(v, bindings={}, symbols=symbols)
            # rl.p(f'{k!r} -> {v!r} -> {r!r}')
            annotations[k] = r
        except NameError:
            msg = (
                f"resolve_type({T.__name__}):"
                f' Cannot resolve names for attribute "{k}" = {v!r}.'
            )
            # msg += f'\n symbols: {symbols}'
            # msg += '\n\n' + indent(traceback.format_exc(), '', '> ')
            # raise NameError(msg) from e
            logger.warning(msg)
            continue
        except TypeError as e:  # pragma: no cover
            msg = f'Cannot resolve type for attribute "{k}".'
            raise ZTypeError(msg) from e
    for f in fields(T):
        assert f.name in annotations
        # msg = f'Cannot get annotation for field {f.name!r}'
        # logger.warning(msg)
        # continue
        f.type = annotations[f.name]


def type_check(type_self: type, k: str, T_expected: type, value_found: object):
    try:
        T_found = type(value_found)
        simple = T_found in [int, float, bool, str, bytes, Decimal, datetime]
        definitely_exclude = T_found in [dict, list, tuple]
        do_it = (not definitely_exclude) and (
            ZuperTypingGlobals.enable_type_checking_difficult or simple
        )
        if do_it:
            # fail = T_found.__name__ != T_expected.__name__ and not isinstance(value_found, T_expected)
            ok = can_be_used_as2(T_found, T_expected)
            if not ok:  # pragma: no cover
                msg = f"The field is not of the expected value"
                # warnings.warn(msg, stacklevel=3)
                raise ZValueError(
                    msg,
                    type_self=type_self,
                    field=k,
                    expected_type=T_expected,
                    found_type=T_found,
                    found_value=value_found,
                    why=ok,
                )
    except TypeError as e:  # pragma: no cover
        msg = f"Cannot judge annotation of {k} (supposedly {value_found}."

        if sys.version_info[:2] == (3, 6):
            # FIXME: warn
            return
        logger.error(msg)
        raise TypeError(msg) from e


def make_type(
    cls: Type[dataclass],
    bindings: Dict[type, type],  # TypeVars
    symbols: Optional[Dict[str, type]] = None,
) -> type:
    from . import debug_print

    assert debug_print
    # from . import debug_print
    #
    # logger.debug(
    #     "make_type: "
    #     + debug_print(dict(cls=cls.__name__, bindings=bindings, symbols=symbols))
    # )
    if symbols is None:
        symbols = {}
    symbols = dict(symbols)

    assert not is_NewType(cls), cls

    if not bindings:
        return cls
    cache_key = (str(cls), str(bindings))
    if cache_enabled:
        if cache_key in MakeTypeCache.cache:
            # print(f'using cached value for {cache_key}')
            return MakeTypeCache.cache[cache_key]

    clsi = get_dataclass_info(cls)
    # generic_att2 = getattr(cls, GENERIC_ATT2, ())
    # prev_bindings = getattr(cls, BINDINGS_ATT, {})
    # assert isinstance(generic_att2, tuple), generic_att2

    # if set(clsi.generic_att) != set(bindings):
    #     msg = "Mismatch with params"
    #     raise ZValueError(msg, di=clsi, bindings=bindings)

    recur = lambda _: replace_typevars(_, bindings=bindings, symbols=symbols)
    new_bindings = {}
    prev_bindings_updated = {k: recur(v) for k, v in clsi.bindings.items()}
    new_bindings.update(prev_bindings_updated)
    new_bindings.update(bindings)

    # orig2 = tuple(recur(v) for v in clsi.orig)
    # orig2 = clsi.orig
    annotations = getattr(cls, "__annotations__", {})
    name_without = get_name_without_brackets(cls.__name__)

    def param_name(x: type) -> str:
        # if x in new_bindings:
        #     tx = new_bindings[x]
        # else:
        #     tx = x
        # x2 = recur(tx)
        x = replace_typevars(x, bindings=new_bindings, symbols=symbols)
        return name_for_type_like(x)

    if clsi.generic_att:
        name2 = "%s[%s]" % (
            name_without,
            ",".join(param_name(_) for _ in clsi.orig),  # generic_att
        )
        # name2 = "%s[%s]" % (name_without, ",".join(param_name(_) for _ in clsi.orig))
    else:
        name2 = name_without

    logger.info(
        "obtained name" + debug_print(dict(cls=cls, bindings=bindings, name2=name2))
    )
    try:
        cls2 = type(name2, (cls,), {"need": lambda: None})
        # logger.info(f'Created class {cls2} ({name2}) and set qualname {cls2.__qualname__}')
    except TypeError as e:  # pragma: no cover
        msg = f'Cannot create derived class "{name2}" from the class.'
        raise ZTypeError(msg, cls=cls) from e

    symbols[name2] = cls2
    symbols[cls.__name__] = cls2  # also MyClass[X] should resolve to the same
    MakeTypeCache.cache[cache_key] = cls2

    class Fake2:
        def __getitem__(self, item):
            n = name_for_type_like(item)
            complete = f"{name_without}[{n}]"
            if complete in symbols:
                return symbols[complete]
            # noinspection PyUnresolvedReferences
            return cls[item]

    if name_without not in symbols:
        symbols[name_without] = Fake2()

    for T, U in bindings.items():
        symbols[T.__name__] = U
        if hasattr(U, "__name__"):
            # dict does not have name
            symbols[U.__name__] = U

    # first of all, replace the bindings in the generic_att

    # generic_att2_new = tuple(recur(_) for _ in generic_att2)

    ts = ()
    for U in bindings.values():
        ts += find_typevars_inside(U)

    generic_att2_new = tuple(unique_preserve_order(ts))

    # logger.debug(
    #     f"creating derived class {name2} with abstract method need() because
    #     generic_att2_new = {generic_att2_new}")

    # rl.p(f'  generic_att2_new: {generic_att2_new}')

    # pprint(f'\n\n{cls.__name__}')
    # pprint(f'binding', bindings=str(bindings))
    # pprint(f'symbols', **symbols)

    new_annotations = {}

    # logger.info(f'annotations ({annotations}) ')
    for k, v0 in annotations.items():
        v = recur(v0)

        # print(f'{v0!r} -> {v!r}')
        if is_ClassVar(v):
            s = get_ClassVar_arg(v)

            if is_Type(s):
                st = get_Type_arg(s)
                concrete = recur(st)
                # logger.info(f'is_Type ({s}) -> {concrete}')
                # concrete = st
                new_annotations[k] = ClassVar[type]
                setattr(cls2, k, concrete)
            else:
                s2 = recur(s)
                new_annotations[k] = ClassVar[s2]
        else:

            new_annotations[k] = v

    # logger.info(f'new_annotations {new_annotations}')
    # pprint('  new annotations', **new_annotations)
    original__post_init__ = getattr(cls, "__post_init__", None)

    if enable_type_checking:

        def __post_init__(self):
            # do it first (because they might change things around)
            if original__post_init__ is not None:
                original__post_init__(self)

            for k, T_expected in new_annotations.items():
                if is_ClassVar(T_expected):
                    continue
                if isinstance(T_expected, type):
                    val = getattr(self, k)
                    type_check(type(self), k=k, value_found=val, T_expected=T_expected)

        # important: do it before dataclass
        setattr(cls2, "__post_init__", __post_init__)

    cls2.__annotations__ = new_annotations

    # logger.info('new annotations: %s' % new_annotations)
    if is_dataclass(cls):
        # note: need to have set new annotations
        # pprint('creating dataclass from %s' % cls2)
        doc = getattr(cls2, "__doc__", None)
        cls2 = dataclass(cls2, unsafe_hash=True)
        setattr(cls2, "__doc__", doc)
    else:
        # noinspection PyUnusedLocal
        def init_placeholder(self, *args, **kwargs):
            if args or kwargs:
                msg = (
                    f"Default constructor of {cls2.__name__} does not know what to do with "
                    f""
                    f""
                    f"arguments."
                )
                msg += f"\nargs: {args!r}\nkwargs: {kwargs!r}"
                msg += f"\nself: {self}"
                msg += f"\nself: {dir(type(self))}"
                msg += f"\nself: {type(self)}"
                raise NoConstructorImplemented(msg)

        if cls.__init__ == object.__init__:
            setattr(cls2, "__init__", init_placeholder)

    cls2.__module__ = cls.__module__
    setattr(cls2, "__name__", name2)

    qn = cls.__qualname__
    qn0, sep, _ = qn.rpartition(".")
    if not sep:
        sep = ""
    setattr(cls2, "__qualname__", qn0 + sep + name2)

    setattr(cls2, BINDINGS_ATT, new_bindings)
    setattr(cls2, GENERIC_ATT2, generic_att2_new)
    # setattr(cls2, GENERIC_BOUND_ATT, clsi.orig)
    setattr(cls2, GENERIC_BOUND_ATT, clsi.orig)

    MakeTypeCache.cache[cache_key] = cls2

    # logger.info(f'started {cls}; hash is {cls.__hash__}')
    # logger.info(f'specialized {cls2}; hash is {cls2.__hash__}')
    # ztinfo("make_type", cls=cls, bindings=bindings, cls2=cls2)
    return cls2


@dataclass
class DataclassInfo:
    generic_att: Tuple[TypeLike, ...]
    bindings: Dict[TypeLike, TypeLike]
    orig: Tuple[TypeLike, ...]


def get_dataclass_info(T: Type[dataclass]) -> DataclassInfo:
    g = getattr(T, GENERIC_ATT2, ())
    b = getattr(T, BINDINGS_ATT, {})
    c = getattr(T, GENERIC_BOUND_ATT, ())
    return DataclassInfo(g, b, c)


def unique_preserve_order(t: Sequence[X]) -> List[X]:
    res = []
    for _ in t:
        if _ not in res:
            res.append(_)
    return res
